# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: labgrid-coordinator.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19labgrid-coordinator.proto\x12\x07labgrid\"\x8a\x01\n\x0f\x43lientInMessage\x12\x1d\n\x04sync\x18\x01 \x01(\x0b\x32\r.labgrid.SyncH\x00\x12\'\n\x07startup\x18\x02 \x01(\x0b\x32\x14.labgrid.StartupDoneH\x00\x12\'\n\tsubscribe\x18\x03 \x01(\x0b\x32\x12.labgrid.SubscribeH\x00\x42\x06\n\x04kind\"\x12\n\x04Sync\x12\n\n\x02id\x18\x01 \x01(\x04\",\n\x0bStartupDone\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"r\n\tSubscribe\x12\x1b\n\x0eis_unsubscribe\x18\x01 \x01(\x08H\x01\x88\x01\x01\x12\x14\n\nall_places\x18\x02 \x01(\x08H\x00\x12\x17\n\rall_resources\x18\x03 \x01(\x08H\x00\x42\x06\n\x04kindB\x11\n\x0f_is_unsubscribe\"g\n\x10\x43lientOutMessage\x12 \n\x04sync\x18\x01 \x01(\x0b\x32\r.labgrid.SyncH\x00\x88\x01\x01\x12(\n\x07updates\x18\x02 \x03(\x0b\x32\x17.labgrid.UpdateResponseB\x07\n\x05_sync\"\xa5\x01\n\x0eUpdateResponse\x12%\n\x08resource\x18\x01 \x01(\x0b\x32\x11.labgrid.ResourceH\x00\x12.\n\x0c\x64\x65l_resource\x18\x02 \x01(\x0b\x32\x16.labgrid.Resource.PathH\x00\x12\x1f\n\x05place\x18\x03 \x01(\x0b\x32\x0e.labgrid.PlaceH\x00\x12\x13\n\tdel_place\x18\x04 \x01(\tH\x00\x42\x06\n\x04kind\"\x9a\x01\n\x11\x45xporterInMessage\x12%\n\x08resource\x18\x01 \x01(\x0b\x32\x11.labgrid.ResourceH\x00\x12\'\n\x07startup\x18\x02 \x01(\x0b\x32\x14.labgrid.StartupDoneH\x00\x12-\n\x08response\x18\x03 \x01(\x0b\x32\x19.labgrid.ExporterResponseH\x00\x42\x06\n\x04kind\"\x9e\x03\n\x08Resource\x12$\n\x04path\x18\x01 \x01(\x0b\x32\x16.labgrid.Resource.Path\x12\x0b\n\x03\x63ls\x18\x02 \x01(\t\x12-\n\x06params\x18\x03 \x03(\x0b\x32\x1d.labgrid.Resource.ParamsEntry\x12+\n\x05\x65xtra\x18\x04 \x03(\x0b\x32\x1c.labgrid.Resource.ExtraEntry\x12\x10\n\x08\x61\x63quired\x18\x05 \x01(\t\x12\r\n\x05\x61vail\x18\x06 \x01(\x08\x1a_\n\x04Path\x12\x1a\n\rexporter_name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x12\n\ngroup_name\x18\x02 \x01(\t\x12\x15\n\rresource_name\x18\x03 \x01(\tB\x10\n\x0e_exporter_name\x1a@\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12 \n\x05value\x18\x02 \x01(\x0b\x32\x11.labgrid.MapValue:\x02\x38\x01\x1a?\n\nExtraEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12 \n\x05value\x18\x02 \x01(\x0b\x32\x11.labgrid.MapValue:\x02\x38\x01\"\x82\x01\n\x08MapValue\x12\x14\n\nbool_value\x18\x01 \x01(\x08H\x00\x12\x13\n\tint_value\x18\x02 \x01(\x03H\x00\x12\x14\n\nuint_value\x18\x03 \x01(\x04H\x00\x12\x15\n\x0b\x66loat_value\x18\x04 \x01(\x01H\x00\x12\x16\n\x0cstring_value\x18\x05 \x01(\tH\x00\x42\x06\n\x04kind\"C\n\x10\x45xporterResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x13\n\x06reason\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_reason\"\x18\n\x05Hello\x12\x0f\n\x07version\x18\x01 \x01(\t\"\x82\x01\n\x12\x45xporterOutMessage\x12\x1f\n\x05hello\x18\x01 \x01(\x0b\x32\x0e.labgrid.HelloH\x00\x12\x43\n\x14set_acquired_request\x18\x02 \x01(\x0b\x32#.labgrid.ExporterSetAcquiredRequestH\x00\x42\x06\n\x04kind\"o\n\x1a\x45xporterSetAcquiredRequest\x12\x12\n\ngroup_name\x18\x01 \x01(\t\x12\x15\n\rresource_name\x18\x02 \x01(\t\x12\x17\n\nplace_name\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\r\n\x0b_place_name\"\x1f\n\x0f\x41\x64\x64PlaceRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x12\n\x10\x41\x64\x64PlaceResponse\"\"\n\x12\x44\x65letePlaceRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x15\n\x13\x44\x65letePlaceResponse\"\x12\n\x10GetPlacesRequest\"3\n\x11GetPlacesResponse\x12\x1e\n\x06places\x18\x01 \x03(\x0b\x32\x0e.labgrid.Place\"\xd2\x02\n\x05Place\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07\x61liases\x18\x02 \x03(\t\x12\x0f\n\x07\x63omment\x18\x03 \x01(\t\x12&\n\x04tags\x18\x04 \x03(\x0b\x32\x18.labgrid.Place.TagsEntry\x12\'\n\x07matches\x18\x05 \x03(\x0b\x32\x16.labgrid.ResourceMatch\x12\x15\n\x08\x61\x63quired\x18\x06 \x01(\tH\x00\x88\x01\x01\x12\x1a\n\x12\x61\x63quired_resources\x18\x07 \x03(\t\x12\x0f\n\x07\x61llowed\x18\x08 \x03(\t\x12\x0f\n\x07\x63reated\x18\t \x01(\x01\x12\x0f\n\x07\x63hanged\x18\n \x01(\x01\x12\x18\n\x0breservation\x18\x0b \x01(\tH\x01\x88\x01\x01\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x0b\n\t_acquiredB\x0e\n\x0c_reservation\"y\n\rResourceMatch\x12\x10\n\x08\x65xporter\x18\x01 \x01(\t\x12\r\n\x05group\x18\x02 \x01(\t\x12\x0b\n\x03\x63ls\x18\x03 \x01(\t\x12\x11\n\x04name\x18\x04 \x01(\tH\x00\x88\x01\x01\x12\x13\n\x06rename\x18\x05 \x01(\tH\x01\x88\x01\x01\x42\x07\n\x05_nameB\t\n\x07_rename\"8\n\x14\x41\x64\x64PlaceAliasRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\"\x17\n\x15\x41\x64\x64PlaceAliasResponse\";\n\x17\x44\x65letePlaceAliasRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\"\x1a\n\x18\x44\x65letePlaceAliasResponse\"\x8b\x01\n\x13SetPlaceTagsRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x34\n\x04tags\x18\x02 \x03(\x0b\x32&.labgrid.SetPlaceTagsRequest.TagsEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x16\n\x14SetPlaceTagsResponse\"<\n\x16SetPlaceCommentRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\"\x19\n\x17SetPlaceCommentResponse\"Z\n\x14\x41\x64\x64PlaceMatchRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x0f\n\x07pattern\x18\x02 \x01(\t\x12\x13\n\x06rename\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_rename\"\x17\n\x15\x41\x64\x64PlaceMatchResponse\"]\n\x17\x44\x65letePlaceMatchRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x0f\n\x07pattern\x18\x02 \x01(\t\x12\x13\n\x06rename\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_rename\"\x1a\n\x18\x44\x65letePlaceMatchResponse\"(\n\x13\x41\x63quirePlaceRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\"\x16\n\x14\x41\x63quirePlaceResponse\"L\n\x13ReleasePlaceRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x15\n\x08\x66romuser\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\x0b\n\t_fromuser\"\x16\n\x14ReleasePlaceResponse\"4\n\x11\x41llowPlaceRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x0c\n\x04user\x18\x02 \x01(\t\"\x14\n\x12\x41llowPlaceResponse\"\xb6\x01\n\x18\x43reateReservationRequest\x12?\n\x07\x66ilters\x18\x01 \x03(\x0b\x32..labgrid.CreateReservationRequest.FiltersEntry\x12\x0c\n\x04prio\x18\x02 \x01(\x01\x1aK\n\x0c\x46iltersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12*\n\x05value\x18\x02 \x01(\x0b\x32\x1b.labgrid.Reservation.Filter:\x02\x38\x01\"F\n\x19\x43reateReservationResponse\x12)\n\x0breservation\x18\x01 \x01(\x0b\x32\x14.labgrid.Reservation\"\xcd\x03\n\x0bReservation\x12\r\n\x05owner\x18\x01 \x01(\t\x12\r\n\x05token\x18\x02 \x01(\t\x12\r\n\x05state\x18\x03 \x01(\x05\x12\x0c\n\x04prio\x18\x04 \x01(\x01\x12\x32\n\x07\x66ilters\x18\x05 \x03(\x0b\x32!.labgrid.Reservation.FiltersEntry\x12:\n\x0b\x61llocations\x18\x06 \x03(\x0b\x32%.labgrid.Reservation.AllocationsEntry\x12\x0f\n\x07\x63reated\x18\x07 \x01(\x01\x12\x0f\n\x07timeout\x18\x08 \x01(\x01\x1ap\n\x06\x46ilter\x12\x37\n\x06\x66ilter\x18\x01 \x03(\x0b\x32\'.labgrid.Reservation.Filter.FilterEntry\x1a-\n\x0b\x46ilterEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1aK\n\x0c\x46iltersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12*\n\x05value\x18\x02 \x01(\x0b\x32\x1b.labgrid.Reservation.Filter:\x02\x38\x01\x1a\x32\n\x10\x41llocationsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\")\n\x18\x43\x61ncelReservationRequest\x12\r\n\x05token\x18\x01 \x01(\t\"\x1b\n\x19\x43\x61ncelReservationResponse\"\'\n\x16PollReservationRequest\x12\r\n\x05token\x18\x01 \x01(\t\"D\n\x17PollReservationResponse\x12)\n\x0breservation\x18\x01 \x01(\x0b\x32\x14.labgrid.Reservation\"E\n\x17GetReservationsResponse\x12*\n\x0creservations\x18\x01 \x03(\x0b\x32\x14.labgrid.Reservation\"\x18\n\x16GetReservationsRequest2\xd2\x0b\n\x0b\x43oordinator\x12I\n\x0c\x43lientStream\x12\x18.labgrid.ClientInMessage\x1a\x19.labgrid.ClientOutMessage\"\x00(\x01\x30\x01\x12O\n\x0e\x45xporterStream\x12\x1a.labgrid.ExporterInMessage\x1a\x1b.labgrid.ExporterOutMessage\"\x00(\x01\x30\x01\x12\x41\n\x08\x41\x64\x64Place\x12\x18.labgrid.AddPlaceRequest\x1a\x19.labgrid.AddPlaceResponse\"\x00\x12J\n\x0b\x44\x65letePlace\x12\x1b.labgrid.DeletePlaceRequest\x1a\x1c.labgrid.DeletePlaceResponse\"\x00\x12\x44\n\tGetPlaces\x12\x19.labgrid.GetPlacesRequest\x1a\x1a.labgrid.GetPlacesResponse\"\x00\x12P\n\rAddPlaceAlias\x12\x1d.labgrid.AddPlaceAliasRequest\x1a\x1e.labgrid.AddPlaceAliasResponse\"\x00\x12Y\n\x10\x44\x65letePlaceAlias\x12 .labgrid.DeletePlaceAliasRequest\x1a!.labgrid.DeletePlaceAliasResponse\"\x00\x12M\n\x0cSetPlaceTags\x12\x1c.labgrid.SetPlaceTagsRequest\x1a\x1d.labgrid.SetPlaceTagsResponse\"\x00\x12V\n\x0fSetPlaceComment\x12\x1f.labgrid.SetPlaceCommentRequest\x1a .labgrid.SetPlaceCommentResponse\"\x00\x12P\n\rAddPlaceMatch\x12\x1d.labgrid.AddPlaceMatchRequest\x1a\x1e.labgrid.AddPlaceMatchResponse\"\x00\x12Y\n\x10\x44\x65letePlaceMatch\x12 .labgrid.DeletePlaceMatchRequest\x1a!.labgrid.DeletePlaceMatchResponse\"\x00\x12M\n\x0c\x41\x63quirePlace\x12\x1c.labgrid.AcquirePlaceRequest\x1a\x1d.labgrid.AcquirePlaceResponse\"\x00\x12M\n\x0cReleasePlace\x12\x1c.labgrid.ReleasePlaceRequest\x1a\x1d.labgrid.ReleasePlaceResponse\"\x00\x12G\n\nAllowPlace\x12\x1a.labgrid.AllowPlaceRequest\x1a\x1b.labgrid.AllowPlaceResponse\"\x00\x12\\\n\x11\x43reateReservation\x12!.labgrid.CreateReservationRequest\x1a\".labgrid.CreateReservationResponse\"\x00\x12\\\n\x11\x43\x61ncelReservation\x12!.labgrid.CancelReservationRequest\x1a\".labgrid.CancelReservationResponse\"\x00\x12V\n\x0fPollReservation\x12\x1f.labgrid.PollReservationRequest\x1a .labgrid.PollReservationResponse\"\x00\x12V\n\x0fGetReservations\x12\x1f.labgrid.GetReservationsRequest\x1a .labgrid.GetReservationsResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'labgrid_coordinator_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_RESOURCE_PARAMSENTRY']._options = None
  _globals['_RESOURCE_PARAMSENTRY']._serialized_options = b'8\001'
  _globals['_RESOURCE_EXTRAENTRY']._options = None
  _globals['_RESOURCE_EXTRAENTRY']._serialized_options = b'8\001'
  _globals['_PLACE_TAGSENTRY']._options = None
  _globals['_PLACE_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_SETPLACETAGSREQUEST_TAGSENTRY']._options = None
  _globals['_SETPLACETAGSREQUEST_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_CREATERESERVATIONREQUEST_FILTERSENTRY']._options = None
  _globals['_CREATERESERVATIONREQUEST_FILTERSENTRY']._serialized_options = b'8\001'
  _globals['_RESERVATION_FILTER_FILTERENTRY']._options = None
  _globals['_RESERVATION_FILTER_FILTERENTRY']._serialized_options = b'8\001'
  _globals['_RESERVATION_FILTERSENTRY']._options = None
  _globals['_RESERVATION_FILTERSENTRY']._serialized_options = b'8\001'
  _globals['_RESERVATION_ALLOCATIONSENTRY']._options = None
  _globals['_RESERVATION_ALLOCATIONSENTRY']._serialized_options = b'8\001'
  _globals['_CLIENTINMESSAGE']._serialized_start=39
  _globals['_CLIENTINMESSAGE']._serialized_end=177
  _globals['_SYNC']._serialized_start=179
  _globals['_SYNC']._serialized_end=197
  _globals['_STARTUPDONE']._serialized_start=199
  _globals['_STARTUPDONE']._serialized_end=243
  _globals['_SUBSCRIBE']._serialized_start=245
  _globals['_SUBSCRIBE']._serialized_end=359
  _globals['_CLIENTOUTMESSAGE']._serialized_start=361
  _globals['_CLIENTOUTMESSAGE']._serialized_end=464
  _globals['_UPDATERESPONSE']._serialized_start=467
  _globals['_UPDATERESPONSE']._serialized_end=632
  _globals['_EXPORTERINMESSAGE']._serialized_start=635
  _globals['_EXPORTERINMESSAGE']._serialized_end=789
  _globals['_RESOURCE']._serialized_start=792
  _globals['_RESOURCE']._serialized_end=1206
  _globals['_RESOURCE_PATH']._serialized_start=980
  _globals['_RESOURCE_PATH']._serialized_end=1075
  _globals['_RESOURCE_PARAMSENTRY']._serialized_start=1077
  _globals['_RESOURCE_PARAMSENTRY']._serialized_end=1141
  _globals['_RESOURCE_EXTRAENTRY']._serialized_start=1143
  _globals['_RESOURCE_EXTRAENTRY']._serialized_end=1206
  _globals['_MAPVALUE']._serialized_start=1209
  _globals['_MAPVALUE']._serialized_end=1339
  _globals['_EXPORTERRESPONSE']._serialized_start=1341
  _globals['_EXPORTERRESPONSE']._serialized_end=1408
  _globals['_HELLO']._serialized_start=1410
  _globals['_HELLO']._serialized_end=1434
  _globals['_EXPORTEROUTMESSAGE']._serialized_start=1437
  _globals['_EXPORTEROUTMESSAGE']._serialized_end=1567
  _globals['_EXPORTERSETACQUIREDREQUEST']._serialized_start=1569
  _globals['_EXPORTERSETACQUIREDREQUEST']._serialized_end=1680
  _globals['_ADDPLACEREQUEST']._serialized_start=1682
  _globals['_ADDPLACEREQUEST']._serialized_end=1713
  _globals['_ADDPLACERESPONSE']._serialized_start=1715
  _globals['_ADDPLACERESPONSE']._serialized_end=1733
  _globals['_DELETEPLACEREQUEST']._serialized_start=1735
  _globals['_DELETEPLACEREQUEST']._serialized_end=1769
  _globals['_DELETEPLACERESPONSE']._serialized_start=1771
  _globals['_DELETEPLACERESPONSE']._serialized_end=1792
  _globals['_GETPLACESREQUEST']._serialized_start=1794
  _globals['_GETPLACESREQUEST']._serialized_end=1812
  _globals['_GETPLACESRESPONSE']._serialized_start=1814
  _globals['_GETPLACESRESPONSE']._serialized_end=1865
  _globals['_PLACE']._serialized_start=1868
  _globals['_PLACE']._serialized_end=2206
  _globals['_PLACE_TAGSENTRY']._serialized_start=2134
  _globals['_PLACE_TAGSENTRY']._serialized_end=2177
  _globals['_RESOURCEMATCH']._serialized_start=2208
  _globals['_RESOURCEMATCH']._serialized_end=2329
  _globals['_ADDPLACEALIASREQUEST']._serialized_start=2331
  _globals['_ADDPLACEALIASREQUEST']._serialized_end=2387
  _globals['_ADDPLACEALIASRESPONSE']._serialized_start=2389
  _globals['_ADDPLACEALIASRESPONSE']._serialized_end=2412
  _globals['_DELETEPLACEALIASREQUEST']._serialized_start=2414
  _globals['_DELETEPLACEALIASREQUEST']._serialized_end=2473
  _globals['_DELETEPLACEALIASRESPONSE']._serialized_start=2475
  _globals['_DELETEPLACEALIASRESPONSE']._serialized_end=2501
  _globals['_SETPLACETAGSREQUEST']._serialized_start=2504
  _globals['_SETPLACETAGSREQUEST']._serialized_end=2643
  _globals['_SETPLACETAGSREQUEST_TAGSENTRY']._serialized_start=2134
  _globals['_SETPLACETAGSREQUEST_TAGSENTRY']._serialized_end=2177
  _globals['_SETPLACETAGSRESPONSE']._serialized_start=2645
  _globals['_SETPLACETAGSRESPONSE']._serialized_end=2667
  _globals['_SETPLACECOMMENTREQUEST']._serialized_start=2669
  _globals['_SETPLACECOMMENTREQUEST']._serialized_end=2729
  _globals['_SETPLACECOMMENTRESPONSE']._serialized_start=2731
  _globals['_SETPLACECOMMENTRESPONSE']._serialized_end=2756
  _globals['_ADDPLACEMATCHREQUEST']._serialized_start=2758
  _globals['_ADDPLACEMATCHREQUEST']._serialized_end=2848
  _globals['_ADDPLACEMATCHRESPONSE']._serialized_start=2850
  _globals['_ADDPLACEMATCHRESPONSE']._serialized_end=2873
  _globals['_DELETEPLACEMATCHREQUEST']._serialized_start=2875
  _globals['_DELETEPLACEMATCHREQUEST']._serialized_end=2968
  _globals['_DELETEPLACEMATCHRESPONSE']._serialized_start=2970
  _globals['_DELETEPLACEMATCHRESPONSE']._serialized_end=2996
  _globals['_ACQUIREPLACEREQUEST']._serialized_start=2998
  _globals['_ACQUIREPLACEREQUEST']._serialized_end=3038
  _globals['_ACQUIREPLACERESPONSE']._serialized_start=3040
  _globals['_ACQUIREPLACERESPONSE']._serialized_end=3062
  _globals['_RELEASEPLACEREQUEST']._serialized_start=3064
  _globals['_RELEASEPLACEREQUEST']._serialized_end=3140
  _globals['_RELEASEPLACERESPONSE']._serialized_start=3142
  _globals['_RELEASEPLACERESPONSE']._serialized_end=3164
  _globals['_ALLOWPLACEREQUEST']._serialized_start=3166
  _globals['_ALLOWPLACEREQUEST']._serialized_end=3218
  _globals['_ALLOWPLACERESPONSE']._serialized_start=3220
  _globals['_ALLOWPLACERESPONSE']._serialized_end=3240
  _globals['_CREATERESERVATIONREQUEST']._serialized_start=3243
  _globals['_CREATERESERVATIONREQUEST']._serialized_end=3425
  _globals['_CREATERESERVATIONREQUEST_FILTERSENTRY']._serialized_start=3350
  _globals['_CREATERESERVATIONREQUEST_FILTERSENTRY']._serialized_end=3425
  _globals['_CREATERESERVATIONRESPONSE']._serialized_start=3427
  _globals['_CREATERESERVATIONRESPONSE']._serialized_end=3497
  _globals['_RESERVATION']._serialized_start=3500
  _globals['_RESERVATION']._serialized_end=3961
  _globals['_RESERVATION_FILTER']._serialized_start=3720
  _globals['_RESERVATION_FILTER']._serialized_end=3832
  _globals['_RESERVATION_FILTER_FILTERENTRY']._serialized_start=3787
  _globals['_RESERVATION_FILTER_FILTERENTRY']._serialized_end=3832
  _globals['_RESERVATION_FILTERSENTRY']._serialized_start=3350
  _globals['_RESERVATION_FILTERSENTRY']._serialized_end=3425
  _globals['_RESERVATION_ALLOCATIONSENTRY']._serialized_start=3911
  _globals['_RESERVATION_ALLOCATIONSENTRY']._serialized_end=3961
  _globals['_CANCELRESERVATIONREQUEST']._serialized_start=3963
  _globals['_CANCELRESERVATIONREQUEST']._serialized_end=4004
  _globals['_CANCELRESERVATIONRESPONSE']._serialized_start=4006
  _globals['_CANCELRESERVATIONRESPONSE']._serialized_end=4033
  _globals['_POLLRESERVATIONREQUEST']._serialized_start=4035
  _globals['_POLLRESERVATIONREQUEST']._serialized_end=4074
  _globals['_POLLRESERVATIONRESPONSE']._serialized_start=4076
  _globals['_POLLRESERVATIONRESPONSE']._serialized_end=4144
  _globals['_GETRESERVATIONSRESPONSE']._serialized_start=4146
  _globals['_GETRESERVATIONSRESPONSE']._serialized_end=4215
  _globals['_GETRESERVATIONSREQUEST']._serialized_start=4217
  _globals['_GETRESERVATIONSREQUEST']._serialized_end=4241
  _globals['_COORDINATOR']._serialized_start=4244
  _globals['_COORDINATOR']._serialized_end=5734
# @@protoc_insertion_point(module_scope)
